/* $Id: child.c,v 2.3 2001/10/23 19:03:27 tony Exp $ */

#include "common.h"
#include <sys/stat.h>
#include <fcntl.h>

#include "child.h"
#include "nntp.h"

static ChildData child_data = CHILD_NULL;

static BOOL child_shouldnt_exit;

static void child_exit(void)
{
    if (child_shouldnt_exit) 
    {
	fprintf(stderr, _("Child process %d exiting unexpectedly\n"),
		child_data.thread_idx);
    }
}

void child_send_command(intercom_Command code, const char *data, ...)
{
    va_list ap;

    va_start(ap, data);
    intercom_send_vcommand(child_data.out_buf, code, child_data.thread_idx,
	    data, ap);
    va_end(ap);
}

static void child_shutdown(BOOL error)
{
    /* Try to let server know we're leaving as gracefully as possible, but
     * don't hang around for response or anything if there's an error */
    deblog_info(_("Thread shutting down %s\n"), error ? _("due to error") :
	    _("naturally"));
    if (child_data.sbuf)
    {
	rw_write_line(child_data.sbuf, N_("QUIT"));
	if (!error)
	{
	    nntp_check_response(child_data.sbuf, _("Bad response to QUIT"),
		    "2");
	}
	rw_close_and_free(child_data.sbuf);
	child_data.sbuf = NULL;
	deblog_info(_("Closed socket OK\n"));
    }
    if (child_data.intercom)
    {
	child_send_command(error ? command_QuietError : command_AllDone,
		N_(""));
    }
    child_shouldnt_exit = FALSE;
}

static void child_connect(const char *comdata)
{
    char dotquad[32];
    int port;
    char *path = XMALLOC(strlen(comdata));

    child_shouldnt_exit = TRUE;
    child_data.pipeline_depth = 0;
    sscanf(comdata, N_("%d %s %s %d"), &port, dotquad, path,
	    &child_data.pipeline);
    child_data.sbuf = nntp_connect(dotquad, port, child_data.thread_idx);
    if (!child_data.sbuf)
    {
	child_shutdown(TRUE);
	return;
    }
    child_send_command(command_Connected, N_(""));
    free(child_data.path);
    if (strcmp(path, N_("!")))
    {
	child_data.path = path;
    }
    else
    {
	free(path);
	child_data.path = NULL;
    }
}

static void child_authenticate(const char *usrnpwd)
{
    char *user, *passwd;

    user = XMALLOC(strlen(usrnpwd) + 1);
    passwd = XMALLOC(strlen(usrnpwd) + 1);
    sscanf(usrnpwd, N_("%s %s"), user, passwd);
    if (nntp_authenticate(child_data.sbuf, user, passwd))
    {
	child_shutdown(TRUE);
    }
    else
    {
	child_send_command(command_Authenticated, N_(""));
    }
    free(user);
    free(passwd);
}

static void child_request_article(void)
{
    child_send_command(command_GetNextArticle, N_(""));
}

static void child_enter_group(const char *data)
{
    char *groupname = XMALLOC(strlen(data) + 1);
    int send_stats;
    long int first, last;
    long int group_ix;

    sscanf(data, N_("%ld %s %d"), &group_ix, groupname, &send_stats);
    deblog_debug(_("Entering group %ld %s, send_stats = %d\n"), group_ix,
	    groupname, send_stats);
    if (nntp_enter_group(child_data.sbuf, groupname, send_stats ? &first :
		NULL, &last))
    {
	deblog_debug(_("Failed to enter %s\n"), groupname);
	child_shutdown(TRUE);
    }
    else if (send_stats)
    {
	deblog_debug(_("%s: %ld-%ld\n"), groupname, first, last);
	child_send_command(command_GroupArticleStats, N_("%ld %ld %ld"),
		group_ix, first, last);
    }
    else
    {
	deblog_debug(_("Entered %s, getting article\n"), groupname);
	child_request_article();
    }
    free(groupname);
}

static void child_issue_xhdr(const char *data)
{
    const char *line;
    long int group_ix;
    char *range = XMALLOC(strlen(data));

    sscanf(data, N_("%ld %s"), &group_ix, range);
    rw_writef(child_data.sbuf, N_("XHDR Message-ID %s\r\n"), range);
    free(range);
    if (!nntp_check_response(child_data.sbuf, _("Bad response to XHDR"), "2"))
    {
	child_shutdown(TRUE);
	return;
    }
    while (line = rw_read_line(child_data.sbuf), line[0] != '.')
    {
	child_send_command(command_GroupContains, N_("%ld %s"), group_ix,
		line);
    }
    child_send_command(command_GroupComplete, N_("%ld"), group_ix);
}

static const char *child_path_filter(const char *line, void *data)
{
    static char *newline = NULL;
    static size_t nllen = 0;
    size_t len;

    UNUSED(data);

    if (child_data.path_done)
	return line;
    if (!line[0])
    {
	child_data.path_done = TRUE;
	return line;
    }
    if (strncmp(line, N_("Path: "), 6))
	return line;
    len = strlen(line) + strlen(child_data.path) + 2;
    if (nllen < len)
	newline = XREALLOC(newline, nllen = len);
    sprintf(newline, N_("Path: %s!%s"), child_data.path, line + 6);
    child_data.path_done = TRUE;
    return newline;
}

/* Returns TRUE if problem */
static BOOL child_pop_article(void)
{
    int fd = -1;
    rw_buf *fbuf = NULL;
    char *filename = NULL;
    BOOL result = FALSE;

    child_data.path_done = FALSE;
    deblog_debug(_("Expecting to receive %d/%d\n"), child_data.group_ix[0],
		child_data.artnum[0]);
    if (!nntp_check_response(child_data.sbuf, _("Bad response to ARTICLE"),
		"2"))
    {
	child_shutdown(TRUE);
	return TRUE;
    }
    /* This may give warning, but I can't find an alternative; it should be
     * safe enough because we only one run at once */
    filename = tempnam(INCOMING_DIR, N_("Mesg"));
    deblog_debug(_("Saving %d/%d to %s\n"), child_data.group_ix[0],
		child_data.artnum[0], filename);
    if (filename)
    {
	fd = creat(filename, 0644);
    }
    if (fd != -1)
    {
	fbuf = rw_create_buf(fd, FALSE, FALSE, -1);
    }
    if (!fbuf || fd == -1 || !filename)
    {
	deblog_critical(
		_("Can't create temporary file to download message: %s\n"),
		strerror(errno));
	child_shutdown(TRUE);
	result = TRUE;
    }
    else if (rw_stof(child_data.sbuf, fbuf, child_data.path ? child_path_filter
		: NULL, NULL) < 0)
    {
	deblog_critical(_("Can't download message to temporary file\n"));
	child_shutdown(TRUE);
	result = TRUE;
    }
    if (fbuf)
	rw_free_buf(fbuf);
    if (fd != -1)
	close(fd);
    if (!result)
    {
	int i;

	child_send_command(command_HaveDownloaded, N_("%d %ld %s"),
		child_data.group_ix[0], child_data.artnum[0], filename);
	for (i = 1; i < child_data.pipeline_depth; ++i)
	{
	    child_data.group_ix[i - 1] = child_data.group_ix[i];
	    child_data.artnum[i - 1] = child_data.artnum[i];
	}
	--child_data.pipeline_depth;
    }
    free(filename);
    return result;
}

static void child_get_article(const char *data)
{
    char *msgid = XMALLOC(strlen(data));

    sscanf(data, N_("%ld %ld %s"),
	    &child_data.group_ix[child_data.pipeline_depth],
	    &child_data.artnum[child_data.pipeline_depth], msgid);
    rw_writef(child_data.sbuf, N_("ARTICLE %s\r\n"), msgid);
    if (++child_data.pipeline_depth >= child_data.pipeline ? PIPELINE_DEPTH : 1)
    {
	child_pop_article();
    }
    child_request_article();
}

static void child_resolve(const char *data)
{
    char *dotquad = nntp_resolve_address(data);

    if (dotquad)
	child_send_command(command_Resolved, dotquad);
    else
	child_shutdown(TRUE);
    free(dotquad);
}

/* Failure to move is fatal */
static void child_move_bad_article(const char *pathname)
{
    char *badname;
    char *baddir = XSTRDUP(pathname);
    char badprefix[] = "XXXXX";
	
    badname = strrchr(baddir, '/');
    *badname = 0;
    badname = strrchr(baddir, '/');
    *badname = 0;
    badname = baddir;
    baddir = XSTRDUPCAT(badname, N_("/failed"));
    free(badname);
    /* This makes tempnam more "thread" safe */
    sprintf(badprefix, "%d", child_data.thread_idx);
    badname = XTEMPNAM(baddir, badprefix);
    deblog_debug(_("Attempting to move bad article to %s\n"), badname);
    if (rename(pathname, badname))
    {
	deblog_critical(_("Couldn't move problem article: %s\n"),
		strerror(errno));
	deblog_error(_("Bad article stuck at %s\n"), pathname);
    }
    deblog_warning(_("Bad article moved to %s\n"), badname);
}

static void child_post_article(const char *pathname)
{
    rw_buf *fbuf;
    int fd;
    const char *response = NULL;
    int result;

    deblog_info(_("Uploading %s\n"), pathname);
    if ((fd = open(pathname, O_RDONLY)) == -1)
    {
	deblog_critical(_("Couldn't open %s for uploading:\n"), pathname);
	deblog_error(N_("%s\n"), strerror(errno));
	return;
    }
    fbuf = rw_create_buf(fd, FALSE, FALSE, -1);
    rw_write_line(child_data.sbuf, N_("POST"));
    if (!nntp_check_response(child_data.sbuf, _("Bad response to POST"), "3"))
    {
	rw_close_and_free(fbuf);
	child_send_command(command_NoUploads, "");
	return;
    }
    result = rw_ftos(fbuf, child_data.sbuf, NULL, NULL);
    rw_close_and_free(fbuf);
    if (!result)
    /* From now on result == 0 means posting was successful, or rejected
     * because it's a duplicate, and response is reason for failure */
    {
	response = rw_read_line(child_data.sbuf);
	deblog_debug(N_("%s\n"), response);
	if (!response)
	{
	    child_send_command(command_Error,
		_("No response from server when posting\n"));
	    return;
	}
	else if (!strncmp(response, N_("441 435 "), 8)
		|| strstr(response, N_("Duplicate"))
		|| strstr(response, N_("Already got")))
	{
	    deblog_message(_("Duplicate upload skipped\n"));
	    child_send_command(command_Duplicate, "");
	}
	else if (response[0] != '2')
	{
	    result = -1;
	}
	else
	{
	    child_send_command(command_Uploaded, "");
	}
    }
    else
    {
	response = strerror(errno);
    }
    if (result)
    {
	deblog_warning(_("Posting failed: %s\n"), response);
	child_move_bad_article(pathname);
	child_send_command(command_BadUpload, "");
    }
    else
    {
	if (unlink(pathname))
	{
	    deblog_critical(_("%s posted OK (or was duplicate),\n"), pathname);
	    deblog_critical(
		    _("but couldn't be removed from outgoing queue:\n"));
	    deblog_error(N_("%s\n"), strerror(errno));
	}
    }
}

static void child_main_loop(void)
{
    const intercom_data *command;

    for (;;)
    {
	command = intercom_get_command(child_data.intercom);
	switch (command->command)
	{
	    case command_Upload:
		child_post_article(command->data);
		break;
	    case command_Resolve:
		child_resolve(command->data);
		break;
	    case command_Connect:
		child_connect(command->data);
		break;
	    case command_ModeReader:
		if (nntp_set_mode(child_data.sbuf, N_("READER")))
		    child_shutdown(TRUE);
		else
		    child_send_command(command_Ready, N_(""));
		break;
	    case command_Auth:
		child_authenticate(command->data);
		break;
	    case command_EnterGroup:
		while (child_data.pipeline_depth)
		    child_pop_article();
		child_enter_group(command->data);
		break;
	    case command_GetArticles:
		child_issue_xhdr(command->data);
		break;
	    case command_GetArticle:
		child_get_article(command->data);
		break;
	    case command_Throttle:
		usleep(atoi(command->data) * 1000);
		child_request_article();
		break;
	    case command_Shutdown:
		deblog_info(
		    _("Received shutdown command, %d %s in pipeline\n"),
		    child_data.pipeline_depth,
		    child_data.pipeline_depth == 1 ? _("article") :
			_("articles"));
		while (child_data.pipeline_depth)
		    child_pop_article();
		child_shutdown(FALSE);
		break;
	    default:
		deblog_error(_("Don't understand command code %d\n"),
			command->command);
		break;
	}
    }
}

void child_start(int thread_idx, int in_fd, int out_fd)
{
    static BOOL exit_handled = FALSE;

    if (!exit_handled)
    {
	atexit(child_exit);
	exit_handled = TRUE;
    }
    child_shouldnt_exit = TRUE;
    child_data.in_buf = rw_create_buf(in_fd, FALSE, TRUE, -1);
    child_data.out_buf = rw_create_buf(out_fd, FALSE, TRUE, -1);
    child_data.intercom = intercom_new(child_data.in_buf);
    child_data.thread_idx = thread_idx;
    child_data.pipeline_depth = 0;
    child_data.path = NULL;
    deblog_use_pipe(child_data.out_buf, thread_idx);
    child_main_loop();
}

