/* Logging for debugging */

/* $Id: deblog.c,v 2.8 2001/10/25 14:40:13 tony Exp $ */

#include "common.h"
#include "intercom.h"
#include "nscurses.h"

static deblog_mask_t deblog_mask = deblog_Mask_Default;

static BOOL deblog_need_lf_flag = FALSE;

static rw_buf *deblog_buf = NULL;
static int deblog_thread_idx;
static BOOL deblog_recursion_allowed = FALSE;
static BOOL deblog_merged_sinks = TRUE;

void deblog_separate_sinks(void)
{
    deblog_merged_sinks = FALSE;
}

void deblog_allow_recursion(BOOL allowed)
{
    deblog_recursion_allowed = allowed;
}

void deblog_vprintf_thread(int thread_idx, deblog_level_t level, const char
	*format, va_list ap)
{
    static BOOL recursion = FALSE;
    BOOL use_std = TRUE;	/* Whether message should go to stdout/stderr;
				   may be used in addition to curses */
    /* FILE *sink = (level == deblog_Level_Message) ? stdout : stderr; */
    FILE *sink = (level == deblog_Level_Info || level ==
	    deblog_Level_Debug) ? stderr : stdout;

    if (recursion && !deblog_recursion_allowed)
    {
	if (nscurses_active())
	{
	    nscurses_emergency_stop();
	}
	fprintf(stderr, _("deblog_vprintf_thread called recursively\n"));
	exit(1);
    }
    if (!(deblog_mask & (1 << level)))
	return;
    recursion = TRUE;
    /* Send message to parent process if we're a thread */
    if (deblog_buf)
    {
	size_t len;
	char *formatted;

	if (vasprintf(&formatted, format, ap) <= 0)
	{
	    fprintf(stderr, _("Unable to send logging message to parent\n"));
	    exit(1);
	}
	/* Replace trailing LF with CR to avoid upsetting rw_read_line */
	len = strlen(formatted) - 1;
	if (formatted[len] == '\n')
	    formatted[len] = '\r';
	intercom_send_command(deblog_buf, command_Log, deblog_thread_idx,
		N_("%d %s"), level, formatted);
	free(formatted);
    }
    else if (nscurses_active())
    {
	/* Errors don't go to curses, instead we shutdown curses and print it
	 * on stdout */
	if (level != deblog_Level_Error)
	{
	    /* Debug/Info don't go to curses if -s option used */
	    if (deblog_merged_sinks || (level != deblog_Level_Debug && level
			!= deblog_Level_Info))
		nscurses_deblog(thread_idx, level, format, ap);
	    /* All messages (additionally) go to stderr if -s used */
	    if (deblog_merged_sinks)
		use_std = FALSE;
	    else
		sink = stderr;
	}
	else
	{
	    nscurses_stop();
	}
    }
    if (!deblog_buf && use_std)
    {
	if (deblog_need_lf_flag && (deblog_merged_sinks || sink == stdout))
	    fputs(N_("\r\n"), sink);
	if (thread_idx != -1)
	{
	    fprintf(sink, _("<%d> "), thread_idx);
	}
	switch (level)
	{
	    case deblog_Level_Error:
		fputs(_("Error: "), sink);
		break;
	    case deblog_Level_Critical:
		fputs(_("Critical: "), sink);
		break;
	    case deblog_Level_Warning:
		fputs(_("Warning: "), sink);
		break;
	    case deblog_Level_Info:
		fputs(_("Info: "), sink);
		break;
	    case deblog_Level_Debug:
		fputs(_("Debug: "), sink);
		break;
	    default:
		break;
	}
	vfprintf(sink, format, ap);
	fflush(sink);
	if (sink == stdout || deblog_merged_sinks)
	    deblog_need_lf_flag = FALSE;
    }
    recursion = FALSE;
    if (level == deblog_Level_Error)
	exit(1);
}

void deblog_printf_thread(int thread_idx, deblog_level_t level, const char
	*format, ...)
{
    va_list ap;
    va_start(ap, format);
    deblog_vprintf_thread(thread_idx, level, format, ap);
    va_end(ap);
}

void deblog_error_thread(int thread_idx, const char *format, ...)
{
    va_list ap;
    va_start(ap, format);
    deblog_vprintf_thread(thread_idx, deblog_Level_Error, format, ap);
    va_end(ap);
}

void deblog_critical_thread(int thread_idx, const char *format, ...)
{
    va_list ap;
    va_start(ap, format);
    deblog_vprintf_thread(thread_idx, deblog_Level_Critical, format, ap);
    va_end(ap);
}

void deblog_warning_thread(int thread_idx, const char *format, ...)
{
    va_list ap;
    va_start(ap, format);
    deblog_vprintf_thread(thread_idx, deblog_Level_Warning, format, ap);
    va_end(ap);
}

void deblog_message_thread(int thread_idx, const char *format, ...)
{
    va_list ap;
    va_start(ap, format);
    deblog_vprintf_thread(thread_idx, deblog_Level_Message, format, ap);
    va_end(ap);
}

void deblog_info_thread(int thread_idx, const char *format, ...)
{
    va_list ap;
    va_start(ap, format);
    deblog_vprintf_thread(thread_idx, deblog_Level_Info, format, ap);
    va_end(ap);
}

void deblog_debug_thread(int thread_idx, const char *format, ...)
{
    va_list ap;
    va_start(ap, format);
    deblog_vprintf_thread(thread_idx, deblog_Level_Debug, format, ap);
    va_end(ap);
}

void deblog_verror_thread(int thread_idx, const char *format, va_list ap)
{
    deblog_vprintf_thread(thread_idx, deblog_Level_Error, format, ap);
}

void deblog_vcritical_thread(int thread_idx, const char *format, va_list ap)
{
    deblog_vprintf_thread(thread_idx, deblog_Level_Critical, format, ap);
}

void deblog_vwarning_thread(int thread_idx, const char *format, va_list ap)
{
    deblog_vprintf_thread(thread_idx, deblog_Level_Warning, format, ap);
}

void deblog_vmessage_thread(int thread_idx, const char *format, va_list ap)
{
    deblog_vprintf_thread(thread_idx, deblog_Level_Message, format, ap);
}

void deblog_vinfo_thread(int thread_idx, const char *format, va_list ap)
{
    deblog_vprintf_thread(thread_idx, deblog_Level_Info, format, ap);
}

void deblog_vdebug_thread(int thread_idx, const char *format, va_list ap)
{
    deblog_vprintf_thread(thread_idx, deblog_Level_Debug, format, ap);
}

void deblog_vprintf(deblog_level_t level, const char *format, va_list ap)
{
    deblog_vprintf_thread(-1, level, format, ap);
}

void deblog_printf(deblog_level_t level, const char *format, ...)
{
    va_list ap;
    va_start(ap, format);
    deblog_vprintf_thread(-1, level, format, ap);
    va_end(ap);
}

void deblog_error(const char *format, ...)
{
    va_list ap;
    va_start(ap, format);
    deblog_vprintf_thread(-1, deblog_Level_Error, format, ap);
    va_end(ap);
}

void deblog_critical(const char *format, ...)
{
    va_list ap;
    va_start(ap, format);
    deblog_vprintf_thread(-1, deblog_Level_Critical, format, ap);
    va_end(ap);
}

void deblog_warning(const char *format, ...)
{
    va_list ap;
    va_start(ap, format);
    deblog_vprintf_thread(-1, deblog_Level_Warning, format, ap);
    va_end(ap);
}

void deblog_message(const char *format, ...)
{
    va_list ap;
    va_start(ap, format);
    deblog_vprintf_thread(-1, deblog_Level_Message, format, ap);
    va_end(ap);
}

void deblog_info(const char *format, ...)
{
    va_list ap;
    va_start(ap, format);
    deblog_vprintf_thread(-1, deblog_Level_Info, format, ap);
    va_end(ap);
}

void deblog_debug(const char *format, ...)
{
    va_list ap;
    va_start(ap, format);
    deblog_vprintf_thread(-1, deblog_Level_Debug, format, ap);
    va_end(ap);
}

void deblog_verror(const char *format, va_list ap)
{
    deblog_vprintf_thread(-1, deblog_Level_Error, format, ap);
}

void deblog_vcritical(const char *format, va_list ap)
{
    deblog_vprintf_thread(-1, deblog_Level_Critical, format, ap);
}

void deblog_vwarning(const char *format, va_list ap)
{
    deblog_vprintf_thread(-1, deblog_Level_Warning, format, ap);
}

void deblog_vmessage(const char *format, va_list ap)
{
    deblog_vprintf_thread(-1, deblog_Level_Message, format, ap);
}

void deblog_vinfo(const char *format, va_list ap)
{
    deblog_vprintf_thread(-1, deblog_Level_Info, format, ap);
}

void deblog_vdebug(const char *format, va_list ap)
{
    deblog_vprintf_thread(-1, deblog_Level_Debug, format, ap);
}

void deblog_need_lf(void)
{
    deblog_need_lf_flag = TRUE;
}

void deblog_set_mask(deblog_mask_t mask)
{
    deblog_mask = mask;
}

void deblog_enable_level(deblog_level_t level)
{
    deblog_mask |= 1 << level;
}

void deblog_disable_level(deblog_level_t level)
{
    deblog_mask &= ~(1 << level);
}

void deblog_use_pipe(rw_buf *buf, int thread_idx)
{
    deblog_buf = buf;
    deblog_thread_idx = thread_idx;
    deblog_debug(_("Ready to log via pipe\n"), thread_idx);
}

