/* $Id: group.c,v 2.2 2001/10/23 19:03:27 tony Exp $ */

#include "common.h"

#include "group.h"
#include "history.h"
#include "stats.h"

GroupData *group_new(const char *name, long int last_downloaded)
{
    GroupData *group = XMALLOC(sizeof(GroupData));

    deblog_info(_("Creating group %s, last downloaded %ld\n"), name,
	    last_downloaded);
    group->name = XSTRDUP(name);
    group->last_downloaded = last_downloaded;
    group->first_available = group->last_available = 0;
    group->last_requested = -1;
    group->status = group_Virginal;
    group->artnums = NULL;
    ++group->interested_threads;
    return group;
}

void group_delete(GroupData *group)
{
    deblog_debug(_("Deleting group %s\n"), group->name);
    free(group->artnums);
    free(group->name);
    free(group);
}

BOOL group_got_stats(GroupData *group, long int first, long int last, const
	char *identifier, int thread_idx)
{
    int i;

    deblog_debug(_("Got stats for %s\n"), group->name);
    group->first_available = first;
    group->last_available = last;
    if (group->last_downloaded >= last)
    {
	group->status = group_AllDone;
	if (group->last_downloaded > last)
	{
	    deblog_warning(_("Last downloaded > last available in %s on %s\n"),
		    group->name, identifier);
	    group->last_downloaded = last;
	}
	else
	{
	    deblog_message_thread(thread_idx, _("No new articles in %s\n"),
		    group->name, identifier);
	}
	return FALSE;
    }
    group->status = group_Downloading;
    if (group->last_downloaded < 0)
	group->last_downloaded = last + group->last_downloaded;
    if (group->last_downloaded < first)
	group->last_downloaded = first - 1;
    group->first_available = group->last_downloaded + 1;
    group->last_requested = group->first_available - 1;
    deblog_message_thread(thread_idx, _("%ld new %s in %s\n"),
	    group->last_available - group->first_available + 1,
	    group->last_available - group->first_available + 1 == 1 ?
		_("article") : _("articles"),
	    group->name);
    group->artnums = XMALLOC((1 + last - group->first_available) *
	    sizeof(group_artnum));
    for (i = 0; i <= last - group->first_available; ++i)
    {
	group->artnums[i].status = artnum_Unknown;
	group->artnums[i].msgid = NULL;
    }
    return TRUE;
}

void group_add_message_id(GroupData *group, long int artnum, const char *msgid)
{
    group_artnum *artinfo = &group->artnums[artnum - group->first_available];
    artinfo->msgid = hashtree_add(the_hashtree, msgid);
    if (artinfo->msgid && artinfo->msgid->status == msgid_Virginal)
    {
	/* deblog_debug(_("%s/%d is available\n"), group->name, artnum); */
	artinfo->status = artnum_Available;
	stats_inc_total_to_fetch();
    }
    else
    {
	deblog_info(_("%s/%d %s is a duplicate\n"), group->name, artnum, msgid);
	artinfo->status = artnum_Duplicate;
    }
}

long int group_get_next_article(GroupData *group, char const **msgid)
{
    long int art;

    deblog_debug(_("Rq for next article to download from %s\n"),
	    group->name);
    deblog_debug(_("Available: %ld-%ld, last requested = %ld\n"),
	    group->first_available, group->last_available,
	    group->last_requested);
    if (group->status != group_Downloading && group->status != group_AllDone)
    {
	deblog_info(_("Rq for article in group %s with no stats\n"),
		group->name);
	return -1;
    }
    else if (group->status == group_AllDone)
    {
	deblog_debug(_("Group %s all done\n"), group->name);
	return 0;
    }
    for (art = group->last_requested + 1 - group->first_available; art <=
	    group->last_available - group->first_available; ++art)
    {
	deblog_debug(_("Looking at article %ld (%ld)\n"), art, art +
		group->first_available);
	switch (group->artnums[art].status)
	{
	    case artnum_Available:
		group->artnums[art].status = artnum_Touched;
		if (group->artnums[art].msgid->status == msgid_Virginal)
		{
		    if (history_check(group->artnums[art].msgid->msgid))
		    {
			deblog_debug(_("%s/%ld %s in history\n"),
				group->name, art + group->first_available,
				group->artnums[art].msgid->msgid);
			group->artnums[art].msgid->status = msgid_Fetched;
			stats_dec_total_to_fetch();
			stats_inc_history_dups();
		    }
		    else
		    {
			deblog_debug(_("Requesting %s/%ld\n"), group->name, art
				+ group->first_available);
			hashtree_change_status(group->artnums[art].msgid,
				msgid_BeingFetched);
			if ((group->last_requested = art +
				    group->first_available) >=
				group->last_available)
			{
			    deblog_debug(_("Last article in %s requested\n"),
				    group->name);
			    group->status = group_AllDone;
			}
			*msgid = group->artnums[art].msgid->msgid;
			return group->last_requested;
		    }
		}
		else
		{
		    deblog_debug(_("%s/%ld %s already requested/downloaded\n"),
			    group->name, art + group->first_available,
			    group->artnums[art].msgid->msgid);
		    stats_dec_total_to_fetch();
		    stats_inc_current_dups();
		}
		break;
	    case artnum_Unknown:
		deblog_debug(_("%s/%ld has unknown status at request\n"),
			group->name, art + group->first_available);
		return -1;
	    case artnum_Unavailable:
		deblog_debug(_("%s/%ld is unavailable for request\n"),
			group->name, art + group->first_available);
		break;
	    case artnum_Duplicate:
		deblog_debug(_("%s/%ld is a duplicate\n"),
			group->name, art + group->first_available);
		break;
	    default:
		deblog_critical(_("Repeat request for %s/%ld\n"),
			group->name, art + group->first_available);
		break;
	}
    }
    if (art + group->first_available > group->last_available)
    {
	deblog_info(_("All articles in %s requested\n"), group->name);
	group->status = group_AllDone;
    }
    return 0;
}

void group_prune_unavailable(GroupData *group)
{
    long int art;

    for (art = 0; art <= group->last_available - group->first_available; ++art)
    {
	if (group->artnums[art].status == artnum_Unknown)
	{
	    deblog_debug(_("%s/%d is unavailable\n"), group->name, art +
		    group->first_available);
	    group->artnums[art].status = artnum_Unavailable;
	}
    }
}

void group_have_downloaded(GroupData *group, long int artnum)
{
    hashtree_change_status(group->artnums[artnum -
	    group->first_available].msgid, msgid_Fetched);
}

void group_write_newsrc(GroupData *group, FILE *fp)
{
    long int last = group->last_downloaded;

    deblog_debug(_("Writing newsrc data for group %s\n"), group->name);
    /* Find last downloaded article */
    if (group->last_requested > 0)
    {
	BOOL found_last = FALSE;

	for (last = 0; last < group->last_available - group->first_available &&
		!found_last; ++last)
	{
	    switch (group->artnums[last].status)
	    {
		case artnum_Unknown:
		case artnum_Available:
		    found_last = TRUE;
		    break;
		case artnum_Unavailable:
		case artnum_Duplicate:
		    /* Skip to next article */
		    break;
		case artnum_Touched:
		    if (group->artnums[last].msgid->status != msgid_Fetched)
			found_last = TRUE;
		    break;
	    }
	}
	last += group->first_available;
    }
    fprintf(fp, "%s %ld\n", group->name, last);
}

