/* It all starts here */

#include "common.h"
#include <sys/stat.h>
#include <fcntl.h>
#include <signal.h>

#include "group.h"
#include "hashtree.h"
#include "histcheck.h"
#include "intercom.h"
#include "kiddy.h"
#include "local.h"
#include "nntp.h"
#include "nscurses.h"
#include "nsthread.h"
#include "server.h"
#include "stats.h"

static enum {
    PM_nntp,
    PM_rnews
} main_post_method;

static int main_nthreads = 0;

/* Allows rnews file to be repaired if something goes wrong in the middle of
 * adding to it */
static off_t main_rnews_size = -1;

static void main_repair_rnews_atexit(void)
{
    static BOOL recursive_exit = FALSE;
    
    deblog_debug(_("Entering main_repair_rnews_atexit\n"));
    if (!recursive_exit && main_rnews_size != -1)
    {
	recursive_exit = TRUE;
	truncate(RNEWS_FILE, main_rnews_size);
	main_rnews_size = -1;
    }
    deblog_debug(_("Leaving main_repair_rnews_atexit\n"));
}

/* For not too harmful signals we can try to shut down neatly */
static void main_signal_handler(int sig)
{
    static BOOL recursive_exit = FALSE;
    
    if (!recursive_exit)
    {
	recursive_exit = TRUE;
	deblog_allow_recursion(TRUE);
	deblog_error(_("Exiting due to signal %d\n"), sig);
    }
    else
	abort();
}

/* For more drastic signals just try to avoid corrupted rnews file at all costs
 */
static void main_drastic_handler(int sig)
{
    static BOOL recursive_exit = FALSE;
    
    if (!recursive_exit)
    {
	recursive_exit = TRUE;
	deblog_allow_recursion(TRUE);
	deblog_critical(_("Aborting due to signal %d\n"), sig);
    }
    deblog_debug(_("Calling main_repair_rnews_atexit due to signal %d\n"), sig);
    main_repair_rnews_atexit();
    if (sig != SIGABRT)
	abort();
}

/* Starts as many new threads as necessary to fill quota; returns TRUE if no
 * more can be started */
static BOOL main_start_new_threads(void)
{
    static BOOL upload_first = FALSE;
    BOOL all_done = TRUE;
    server_GFT_Result sgr = server_GFT_OK;

    while (main_nthreads < maincf_get_max_threads() && sgr == server_GFT_OK)
    {
	ServerData *serv;

	switch (sgr = server_get_for_threading(&serv))
	{
	    case server_GFT_OK:
		all_done = FALSE;
		if (nsthread_new(serv, upload_first))
		{
		    ++serv->nthreads;
		    ++main_nthreads;
		    upload_first = !upload_first;
		}
		else
		    return FALSE;
	    case server_GFT_AllBusy:
		all_done = FALSE;
		break;
	    default:
		break;
	}
    }
    return all_done;
}

static void main_got_message_id(NSThread *thread, const char *data)
{
    long int group_ix;
    long int artnum;
    char *msgid = XMALLOC(strlen(data));

    sscanf(data, N_("%ld %ld %s"), &group_ix, &artnum, msgid);
    group_add_message_id(thread->serv->group[group_ix], artnum, msgid);
    free(msgid);
}

/* How long a child should throttle for if it overtakes list of available
 * articles */
#ifndef UNKNOWN_THROTTLE
#define UNKNOWN_THROTTLE 1000
#endif

static void main_shutdown_or_upload(NSThread *thread)
{
    deblog_message_thread(thread->thread_idx, _("Download complete\n"));
    if (thread->serv->nuploaded < thread->serv->noutgoing)
    {
	deblog_message_thread(thread->thread_idx, _("Uploading articles\n"));
	nsthread_send_command(thread, command_Upload,
		thread->serv->outgoing[thread->serv->nuploaded++]);
    }
    else
    {
	deblog_info_thread(thread->thread_idx, _("Shutting down\n"));
	nsthread_send_command(thread, command_Shutdown, N_(""));
    }
}

static void main_get_next_article(NSThread *thread)
{
    long int art;
    const char *msgid;

    deblog_debug_thread(thread->thread_idx, _("Another article wanted\n"));
    /* If server's group_ix is out of range, it means another thread has just
     * downloaded last message, so shut down this thread too */ 
    if (thread->serv->group_ix >= thread->serv->ngroups ||
	    thread->serv->group_ix == -1)
    {
	deblog_debug_thread(thread->thread_idx,
		_("No more articles available to fetch\n"));
	main_shutdown_or_upload(thread);
	return;
    }
    deblog_debug_thread(thread->thread_idx,
	    _("Current group for fetching articles is %s\n"),
	    thread->serv->group[thread->serv->group_ix]->name);
    art = group_get_next_article(thread->serv->group[thread->serv->group_ix],
		&msgid);
    deblog_debug_thread(thread->thread_idx,
	    _("Next article from group %s is %ld\n"),
	    thread->serv->group[thread->serv->group_ix]->name, art);
    switch (art)
    {
	case 0:
	    if (server_get_next_group_articles(thread->serv) == -1)
	    {
		thread->serv->status = server_AllDone;
		main_shutdown_or_upload(thread);
	    }
	    else
	    {
		deblog_info_thread(thread->thread_idx,
			_("Article requests now from %s\n"),
			thread->serv->group[thread->serv->group_ix]->name);
		main_get_next_article(thread);
	    }
	    break;
	case -1:
	    deblog_info_thread(thread->thread_idx,
		    _("Article requests caught up with data, throttling\n"));
	    nsthread_send_command(thread, command_Throttle, N_("%d"),
		    UNKNOWN_THROTTLE);
	    break;
	default:
	    deblog_debug_thread(thread->thread_idx,
		    _("Ordering to get article\n"));
	    nsthread_send_command(thread, command_GetArticle, N_("%ld %ld %s"),
		    thread->serv->group_ix, art, msgid);
	    break;
    }
}

static void main_group_complete(NSThread *thread, const char *data, BOOL prune)
{
    long int group_ix;

    sscanf(data, N_("%ld"), &group_ix);
    if (prune)
	group_prune_unavailable(thread->serv->group[group_ix]);
    group_ix = server_get_next_group_stats(thread->serv);
    deblog_info_thread(thread->thread_idx, _("Now entering group %ld\n"),
	    group_ix);
    if (group_ix != -1)
    {
	nsthread_send_command(thread, command_EnterGroup, N_("%d %s %d"),
		group_ix, thread->serv->group[group_ix]->name, TRUE);
    }
    else
    {
	deblog_message_thread(thread->thread_idx, _("Downloading articles\n"));
	main_get_next_article(thread);
    }
}

static void main_got_group_stats(NSThread *thread, const char *data)
{
    long int group_ix, first, last;
    GroupData *group;

    sscanf(data, N_("%ld %ld %ld"), &group_ix, &first, &last);
    group = thread->serv->group[group_ix];
    if (group_got_stats(group, first, last, thread->serv->identifier,
		thread->thread_idx))
    {
	nsthread_send_command(thread, command_GetArticles, N_("%ld %d-%d"),
		group_ix, group->first_available, last);
    }
    else
    {
	main_group_complete(thread, data, FALSE);
    }
}

/* Adds article read by in_fb to rnews batch file; uses filename for message
 * and to read size with stat, and to delete it */
static void main_hdl_rnews(rw_buf *in_fb, const char *filename)
{
    struct stat fdata;
    int fd;
    static rw_buf *out_fb = NULL;
    BOOL result = FALSE;

    if (!out_fb)
    {
	deblog_info(_("Opening rnews file %s\n"), RNEWS_FILE);
	result = ((fd = open(RNEWS_FILE, O_WRONLY | O_CREAT | O_APPEND, 0644))
		< 0);
	if (!result)
	    result = ((out_fb = rw_create_buf(fd, FALSE, FALSE, -1)) == NULL);
	if (result)
	    deblog_error(_("Couldn't open rnews file: %s\n"), strerror(errno));
    }
    /* Remember size of rnews file in case copy fails and it needs repairing */
    main_rnews_size = stat(RNEWS_FILE, &fdata) ? 0 : fdata.st_size;
    deblog_debug(_("Adding %s to rnews file\n"), filename);
    if (!result)
	result = (stat(filename, &fdata) != 0);
    if (!result)
	result = ((fd = open(filename, O_RDONLY)) < 0);
    if (!result)
    {
	result = (rw_writef(out_fb, N_("#! rnews %08d\n"), fdata.st_size) <= 0);
    }
    if (!result)
	result = (rw_copy(in_fb, out_fb, NULL, NULL) != 0);
    if (result)
    {
	truncate(RNEWS_FILE, main_rnews_size);
	main_rnews_size = -1;
	deblog_error(_("Couldn't add article to rnews file: %s\n"),
		strerror(errno));
    }
    main_rnews_size = -1;
    remove(filename);
}

/* If INN has already seen a Message_ID with CHECK, we must use TAKETHIS to
 * post it, otherwise IHAVE */
static BOOL main_use_takethis = FALSE;

/* Posts article read by fbuf; returns TRUE if rejected */
static BOOL main_hdl_ihave(rw_buf *fbuf, const char *msgid, const char
	*filename)
{
    rw_writef(local_buf, N_("%s %s\r\n"),
	    main_use_takethis ? N_("TAKETHIS") : N_("IHAVE"), msgid);
    if (!main_use_takethis && !nntp_check_response(local_buf,
		_("Article rejected before sending"), "3"))
    {
	return TRUE;
    }
    if (rw_ftos(fbuf, local_buf, NULL, NULL))
    {
	deblog_error(_("Couldn't post article to local server: %s\n"),
		strerror(errno));
    }
    remove(filename);
    return nntp_check_response(local_buf,
	    _("Article rejected after sending"), "2") == NULL;
}

static void main_have_downloaded(NSThread *thread, const char *data)
{
    long int group_ix;
    long int artnum;
    int fd;
    rw_buf *in_fb = NULL;
    BOOL result = FALSE;
    char *filename = XMALLOC(strlen(data));

    sscanf(data, N_("%ld %ld %s"), &group_ix, &artnum, filename);
    result = ((fd = open(filename, O_RDONLY)) < 0);
    if (!result)
	result = ((in_fb = rw_create_buf(fd, FALSE, FALSE, -1)) == NULL);
    if (result)
	deblog_error(_("Couldn't open article file: %s\n"), strerror(errno));
    switch (main_post_method)
    {
	case PM_nntp:
	    main_hdl_ihave(in_fb, thread->serv->group[group_ix]->
		artnums[artnum - thread->serv->group[group_ix]->first_available]
		    .msgid->msgid,
		    filename);
	    break;
	case PM_rnews:
	    main_hdl_rnews(in_fb, filename);
	    break;
    }
    rw_close_and_free(in_fb);
    free(filename);
    /* Ignore result, behave as if fetch was successful */
    group_have_downloaded(thread->serv->group[group_ix], artnum);
    stats_inc_fetched();
}

static void main_start_fetching(NSThread *thread)
{
    long int group_ix = server_get_next_group_stats(thread->serv);

    deblog_debug_thread(thread->thread_idx,
	    _("Ready to fetch, entering group %ld\n"), group_ix);
    if (group_ix == -1)
    {
	group_ix = 0;
	if (thread->serv->enter_group)
	{
	    nsthread_send_command(thread, command_EnterGroup,
		    N_("%d %s %d"), group_ix,
		    thread->serv->group[group_ix]->name, FALSE);
	}
	else
	{
	    deblog_message_thread(thread->thread_idx,
		    _("Downloading articles\n"));
	    main_get_next_article(thread);
	}
    }
    else
    {
	nsthread_send_command(thread, command_EnterGroup,
		N_("%d %s %d"), group_ix,
		thread->serv->group[group_ix]->name, TRUE);
    }
}

static void main_upload_next(NSThread *thread)
{
    if (thread->serv->noutgoing > thread->serv->nuploaded)
    {
	nsthread_send_command(thread, command_Upload,
		thread->serv->outgoing[thread->serv->nuploaded++]);
    }
    else
    {
	deblog_message_thread(thread->thread_idx, _("Uploads completed\n"));
	if (thread->upload_first)
	{
	    main_start_fetching(thread);
	}
	else
	{
	    deblog_info_thread(thread->thread_idx, _("Shutting down\n"));
	    nsthread_send_command(thread, command_Shutdown, N_(""));
	}
    }
}

static void main_loop(void)
{
    const intercom_data *command;
    NSThread *thread;
    struct timeval tv;
    long int bytes;

    for (;;)
    {
	command = kiddy_get_command();
	thread = nsthread_lookup(command->thread_idx);
	/*main_check_available(thread);*/
	switch (command->command)
	{
	    case command_Resolved:
		thread->serv->dotquad = XSTRDUP(command->data);
		thread->serv->status = server_Resolved;
		deblog_message_thread(thread->thread_idx,
			_("Connecting to %s\n"), thread->serv->identifier);
		nsthread_send_command(thread, command_Connect, N_("%d %s %s"), 
			thread->serv->port, thread->serv->dotquad,
			thread->serv->path ? thread->serv->path : N_("!"),
			thread->serv->pipeline);
		thread->serv->status = server_Connected;
		/* Try to start more threads in case we were blocked on missing
		 * info */
		main_start_new_threads();
		break;
	    case command_Log:
		deblog_printf_thread(command->thread_idx, atoi(command->data),
			strchr(command->data, ' ') + 1);
		break;
	    case command_Connected:
		if (thread->serv->user)
		{
		    nsthread_send_command(thread, command_LogOn, N_("%s %s"),
			    thread->serv->user, thread->serv->password);
		    break;
		}
		/* else fall through */
	    case command_Authenticated:
		if (thread->serv->mode_reader)
		{
		    nsthread_send_command(thread, command_ModeReader, N_(""));
		    break;
		}
		/* else fall through */
	    case command_Ready:
		if ((thread->upload_first && thread->serv->noutgoing >
			thread->serv->nuploaded)
		    || !thread->serv->ngroups)
		{
		    deblog_message_thread(thread->thread_idx,
			    _("Uploading articles\n"));
		    nsthread_send_command(thread, command_Upload,
			    thread->serv->outgoing[thread->serv->nuploaded++]);
		}
		else
		{
		    main_start_fetching(thread);
		}
		break;
	    case command_GroupArticleStats:
		main_got_group_stats(thread, command->data);
		/* Try to start more threads in case we were blocked on missing
		 * info */
		main_start_new_threads();
		break;
	    case command_GroupContains:
		main_got_message_id(thread, command->data);
		break;
	    case command_GroupComplete:
		main_group_complete(thread, command->data, TRUE);
		break;
	    case command_GetNextArticle:
		main_get_next_article(thread);
		break;
	    case command_HaveDownloaded:
		main_have_downloaded(thread, command->data);
		break;
	    case command_Error:
		deblog_critical(command->data);
		/* NB fall through to next clause */
	    case command_QuietError:
		thread->serv->status = server_Error;
		/* NB fall through to next clause */
	    case command_AllDone:
		deblog_debug(_("Thread %d finished\n"), command->thread_idx);
		nsthread_delete(thread);
		--main_nthreads;
		if (main_start_new_threads() && ! main_nthreads)
		{
		    return;
		}
		break;
	    case command_AmountSent:
		sscanf(command->data, "%ld %ld %ld", &bytes, &tv.tv_sec,
			&tv.tv_usec);
		stats_update_bytes_sent(bytes, &tv);
		break;
	    case command_AmountRcvd:
		sscanf(command->data, "%ld %ld %ld", &bytes, &tv.tv_sec,
			&tv.tv_usec);
		stats_update_bytes_rcvd(bytes, &tv);
		break;
	    case command_Uploaded:
		stats_inc_uploaded();
		main_upload_next(thread);
		break;
	    case command_Duplicate:
		stats_dec_total_to_upload();
		stats_inc_duplicates();
		main_upload_next(thread);
		break;
	    case command_BadUpload:
		stats_dec_total_to_upload();
		stats_inc_bad_uploads();
		main_upload_next(thread);
		break;
	    case command_NoUploads:
		thread->serv->nuploaded = thread->serv->noutgoing;
		main_upload_next(thread);
		break;
	    default:
		deblog_error(_("Don't understand command code %d\n"),
			command->command);
		break;
	}
    }
}

int main(int argc, char **argv)
{
    int arg;
    BOOL use_curses = FALSE;
    
    if (argc > 4)
	deblog_warning(_("Too many arguments\n"));
    for (arg = 1; arg < argc; ++arg)
    {
	if (!strcmp(argv[arg], "-v") || !strcmp(argv[arg], "-vv"))
	{
	    deblog_enable_level(deblog_Level_Info);
	    if (argv[arg][2] == 'v')
		deblog_enable_level(deblog_Level_Debug);
	}
	else if (!strcmp(argv[arg], "-s"))
	{
	    deblog_separate_sinks();
	}
	else if (!strcmp(argv[arg], "-f"))
	{
#ifdef HAVE_CURSES_H
	    use_curses = TRUE;
#else
	    deblog_warning(_("Full-screen option (-f) not supported\n"));
#endif
	}
	else
	{
	    deblog_warning(_("Unknown option '%s'\n"), argv[arg]);
	}
    }
    deblog_message(_("%s version %s\n"), PACKAGE, VERSION);
    local_config_init();
    maincf_init();
    if (maincf_get_history_method())
    {
	main_use_takethis = FALSE;
	if (!strcmp(maincf_get_history_method(), N_("check")))
	{
	    histcheck_init();
	    main_use_takethis = TRUE;
	    if (nntp_set_mode(local_buf, N_("STREAM")))
	    {
		deblog_error(
		    _("Unable to set local connection to streaming mode\n"));
	    }
	}
	else if (strcmp(maincf_get_history_method(), N_("none")))
	    deblog_error(_("history_method '%s' not known\n"));
    }
    if (maincf_get_post_method())
    {
	if (!strcmp(maincf_get_post_method(), N_("nntp")))
	{
	    main_post_method = PM_nntp;
	    local_init();
	}
	else if (!strcmp(maincf_get_post_method(), N_("rnews")))
	{
	    main_post_method = PM_rnews;
	}
	else
	{
	    deblog_error(_("post_method '%s' not known\n"));
	}
    }
    kiddy_spawn_all();
    atexit(main_repair_rnews_atexit);
    signal(SIGHUP, main_signal_handler);
    signal(SIGINT, main_signal_handler);
    signal(SIGPIPE, main_signal_handler);
    signal(SIGTERM, main_signal_handler);
    signal(SIGQUIT, main_drastic_handler);
    signal(SIGILL, main_drastic_handler);
    signal(SIGFPE, main_drastic_handler);
    signal(SIGSEGV, main_drastic_handler);
    signal(SIGABRT, main_drastic_handler);
    signal(SIGALRM, SIG_IGN);
    signal(SIGUSR1, SIG_IGN);
    signal(SIGUSR2, SIG_IGN);
    server_scan_rc_dir();
    the_hashtree = hashtree_new();
    stats_start();
    if (use_curses)
	nscurses_start(maincf_get_max_threads());
    if (main_start_new_threads())
	return 0;
    main_loop();
    stats_display();
    deblog_message(_("News fetch complete\n"));
    if (use_curses && nscurses_active())
	nscurses_stop();
    stats_summary();

    return 0;
}
