/* $Id: nntp.c,v 2.2 2001/09/16 21:24:00 tony Exp $ */

#include "common.h"

#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>

#include "nntp.h"

char *nntp_resolve_address(const char *name)
{
    struct hostent *hinfo;
    char **aliases, **addrs;
    
    hinfo = gethostbyname(name);
    if (!hinfo || hinfo->h_addrtype != AF_INET)
    {
	deblog_critical(_("Can't resolve address \"%s\"\n"),
	    name);
	return NULL;
    }
    deblog_info(_("Master host name: %s\n"), hinfo->h_name);
    deblog_info(_("Aliases:\n"));
    for (aliases = hinfo->h_aliases; *aliases; ++aliases)
    {
	deblog_info(N_("\t%s\n"), *aliases);
    }
    deblog_info(_("Addresses:\n"));
    for (addrs = hinfo->h_addr_list; *addrs; ++addrs)
    {
	deblog_info(_("\t%s\n"), inet_ntoa(*(struct in_addr *) *addrs));
    }
    return XSTRDUP(inet_ntoa(*(struct in_addr *) hinfo->h_addr_list[0]));
}

const char *nntp_check_response(rw_buf *sbuf, const char *message, const char
	*allow)
{
    char digit[2] = " ";
    const char *response = rw_read_line(sbuf);
    
    if (response)
	digit[0] = response[0];
    if (!response || !strpbrk(digit, allow))
    {
	deblog_critical(N_("%s: %s\n"), message,
		response ? response : _("no response from server"));
	return NULL;
    }
    else
    {
	deblog_debug(N_("%s\n"), response);
    }
    return response;
}

rw_buf *nntp_connect(const char *dotquad, int port, int thread_idx)
{
    int sock;
    rw_buf *buf;

    deblog_info(_("Connecting to %s:%d\n"), dotquad, ntohs(port));
    sock = socket(AF_INET, SOCK_STREAM, 0);
    if (sock != -1)
    {
	struct sockaddr_in addr;

	addr.sin_family = AF_INET;
	addr.sin_port = port;
	inet_aton(dotquad, &addr.sin_addr);
	if (connect(sock, (struct sockaddr *) &addr, sizeof(addr))
		== -1)
	{
	    int terrno = errno;

	    close(sock);
	    errno = terrno;
	    sock = -1;
	}
    }
    else
    {
	deblog_info(_("Couldn't create socket\n"));
    }
    if (sock == -1)
    {
	deblog_critical(_("Can't connect to %s: %s\n"),
		dotquad, strerror(errno));
	return NULL;
    }
    buf = rw_create_buf(sock, TRUE, FALSE, thread_idx);
    if (!nntp_check_response(buf, _("Bad/no response when connecting"), "2"))
    {
	close(sock);
	return NULL;
    }
    return buf;
}

BOOL nntp_authenticate(rw_buf *buf, const char *user, const char *passwd)
{
    deblog_info(_("Authenticating user %s\n"), user);
    rw_writef(buf, N_("AUTHINFO USER %s\r\n"), user);
    if (!nntp_check_response(buf, _("User name rejected"), "23"))
    {
	return TRUE;
    }
    rw_writef(buf, N_("AUTHINFO PASS %s\r\n"), passwd);
    return (nntp_check_response(buf, _("Password rejected"), "2") == NULL);
}

BOOL nntp_set_mode(rw_buf *buf, const char *mode)
{
    deblog_info(_("Sending MODE %s\n"), mode);
    rw_writef(buf, N_("MODE %s\r\n"), mode);
    return (nntp_check_response(buf, _("Bad response to MODE command"), "2")
	    == NULL);
}

BOOL nntp_enter_group(rw_buf *buf, const char *groupname, long int *first, long
	int *last)
{
    const char *response;

    deblog_info(_("Entering group %s\n"), groupname);
    rw_writef(buf, N_("GROUP %s\r\n"), groupname);
    if ((response = nntp_check_response(buf, _("Can't enter group"), "2")) ==
	    NULL)
    {
	return TRUE;
    }
    if (first && last)
    {
	sscanf(response, N_("%*d %*d %ld %ld"), first, last);
    }
    return FALSE;
}

