/* Routines for reading and writing files etc */

/* $Id: rw.c,v 2.1 2001/08/12 17:23:39 tony Exp $ */

#include "common.h"

#include "child.h"

#define BUFLEN 80

/* Minimum number of bytes to read into at end of buffer */
#define MINREAD 16

rw_buf *rw_create_buf(int fd, BOOL crlf, BOOL threadsafe, int thread_idx)
{
    rw_buf *buf;
    
    buf = XMALLOC(sizeof(rw_buf));
    buf->buf = XMALLOC(buf->buflen = BUFLEN);
    buf->buf[0] = 0;
    buf->fd = fd;
    buf->surplus_start = buf->surplus_size = 0;
    buf->status = rw_Normal;
    buf->crlf = crlf;
    buf->threadsafe = threadsafe;
    buf->thread_idx = thread_idx;
    buf->rcvd = buf->sent = 0;
    if (buf->thread_idx != -1)
    {
	struct timezone tz;
	
	gettimeofday(&buf->tv_read, &tz);
	buf->tv_write = buf->tv_read;
    }
    return buf;
}

void rw_free_buf(rw_buf *buf)
{
    if (buf)
	free(buf->buf);
    free(buf);
}

int rw_close(rw_buf *buf)
{
    int result = 0;

    if (!buf)
	return -1;
    if (buf->fd > -1)
    {
	result = close(buf->fd);
	buf->fd = -1;
    }
    else
    {
	deblog_warning(_("Closing already closed rw_buf file descriptor\n"));
    }
    return result;
}

int rw_close_and_free(rw_buf *buf)
{
    int result = rw_close(buf);

    rw_free_buf(buf);
    return result;
}

/* Size of terminator used by a particular rw_buf */
static size_t rw_term_size(const rw_buf *buf)
{
    return buf->crlf ? 2 : 1;
}

/* If a string contains a terminator within len, returns its offset,
 * otherwise -1 
 */
static int rw_str_term(const rw_buf *buf, const char *str, size_t len)
{
    size_t i;
    for (i = 0; i < len; ++i)
    {
	if (str[i] == '\n')
        {
            if (buf->crlf)
            {
                if (str[i - 1] == '\r')
                    return i - 1;
            }
            else
	        return i;
        }
    }
    return -1;
}

/* TRANSFER_STATS_INTERVAL is how often (in ms) stats on bytes transferred per
 * thread should be sent to parent */
#ifdef TRANSFER_STATS_INTERVAL
#define INTERVAL_TV \
 { (TRANSFER_STATS_INTERVAL) / 1000, ((TRANSFER_STATS_INTERVAL) * 1000) % 1000 }
#else
#define INTERVAL_TV { 0, 200000 }
#endif
static struct timeval rw_interval = INTERVAL_TV;

static void rw_send_stats(intercom_Command cmd, struct timeval *last_time, long
	int *amount)
{
    struct timeval now;
    struct timeval difference;
    struct timezone tz;

    gettimeofday(&now, &tz);
    /* deblog_debug(_("rw_send_stats %ld at %.2f\n"), *amount, FLOAT_TV(&now));
     */
    /* Subtract last_time from now */
    xtimeval_subtract(&now, last_time, &difference);
    if (timercmp(&difference, &rw_interval, > ))
    {
	child_send_command(cmd, "%ld %ld %ld", *amount, last_time->tv_sec,
		last_time->tv_usec);
	*last_time = now;
	*amount = 0;
    }
}

const char *rw_read_line(rw_buf *buf)
{
    if (buf->status)
	return NULL;
    if (buf->surplus_size)
    {
	memmove(buf->buf, buf->buf + buf->surplus_start, buf->surplus_size);
	buf->surplus_start = 0;
    }
    for (;;)
    {
        int n;
        int toff = rw_str_term(buf, buf->buf, buf->surplus_size);

        if (toff != -1)
        {
	    buf->buf[toff] = 0;
    	    buf->surplus_start = toff + rw_term_size(buf);
	    buf->surplus_size -= toff + rw_term_size(buf);
	    return buf->buf;
        }
        if (buf->buflen - buf->surplus_size < MINREAD)
        {
	    buf->buf = XREALLOC(buf->buf, buf->buflen += BUFLEN);
	    /*
	    if (!buf->buf)
	    {
	        buf->status = rw_Err;
	        return NULL;
	    }
	    */
        }
        n = read(buf->fd, buf->buf + buf->surplus_size, 
                buf->buflen - buf->surplus_size);
	if (!n)
	{
	    buf->status = rw_Eof;
	    buf->buf[buf->surplus_size] = 0;
	    buf->surplus_start = buf->surplus_size = 0;
	    return buf->surplus_size ? buf->buf : NULL;
	}
	if (n < 0)
	{
	    buf->status = rw_Err;
	    return NULL;
	}
	buf->surplus_size += n;
	if (buf->thread_idx != -1)
	{
	    buf->rcvd += n;
	    rw_send_stats(command_AmountRcvd, &buf->tv_read, &buf->rcvd);
	}
    }
    return NULL;
}

int rw_buf_eof(const rw_buf *buf)
{
    return (buf->status == rw_Eof);
}

int rw_write_line(rw_buf *buf, const char *str)
{
    int result;
    char *termstr = XSTRDUPCAT(str, buf->crlf ? N_("\r\n") : N_("\n"));

    result = rw_write(buf, termstr);
    free(termstr);
    return result;
}

int rw_write(rw_buf *buf, const char *str)
{
    const char *s = str;
    unsigned int l = strlen(str);
    int n;

    do 
    {
        n = write(buf->fd, s, l);
	if (n < 0)
	{
	    return n;
	}
	l -= n;
	s += n;
	if (buf->thread_idx != -1)
	{
	    buf->sent += n;
	}
	if (l > 0 && buf->threadsafe)
	{
	    /* Write this directly to stderr, because we'll almost certainly be
	     * in a subthread and can no longer trust deblog */
	    fputs(_("\nError: Unable to write complete line to pipe\n"),
		    stderr);
	    exit(1);
	}
    } while (l > 0);
    if (buf->thread_idx != -1)
    {
	rw_send_stats(command_AmountSent, &buf->tv_write, &buf->sent);
    }
    return n;
}

int rw_ftos(rw_buf *src, rw_buf *dest, rw_modifier modifier, void *handle)
{
    const char *ln;

    for (ln = rw_read_line(src); ln; ln = rw_read_line(src))
    {
        if (modifier)
            ln = modifier(ln, handle);
        if (ln && ln[0] == '.')
        {
            int n;

            do 
            {
                char esc = '.';

                n = write(dest->fd, &esc, 1);
                if (n < 0)
                    return n;
            } while (!n);
        }
        if (ln && rw_write_line(dest, ln) < 0)
            return -1;
    }
    if (!rw_buf_eof(src))
        return -1;
    /* Should write 3 bytes including EOL */
    return rw_write_line(dest, N_(".")) == 3 ? 0 : -1;
}

int rw_stof(rw_buf *src, rw_buf *dest, rw_modifier modifier, void *handle)
{
    const char *ln;

    for (ln = rw_read_line(src); ln; ln = rw_read_line(src))
    {
        const char *escln = ln;

        if (ln[0] == '.')
        {
            if (!ln[1])
                return 0;
            ++escln;
        }
        if (modifier)
            escln = modifier(escln, handle);
        if (escln && rw_write_line(dest, escln) < 0)
            return -1;
    }
    /* Shouldn't get EOF */
    return -1;
}

int rw_copy(rw_buf *src, rw_buf *dest, rw_modifier modifier, void *handle)
{
    const char *ln;

    for (ln = rw_read_line(src); ln; ln = rw_read_line(src))
    {
        const char *escln = ln;

        if (modifier)
            escln = modifier(escln, handle);
        if (escln && rw_write_line(dest, escln) < 0)
            return -1;
    }
    /* Assume EOF */
    return 0;
}

int rw_writef(rw_buf *buf, const char *format, ...)
{
    va_list ap;
    int result;

    va_start(ap, format);
    result = rw_vwritef(buf, format, ap);
    va_end(ap);
    return result;
}

int rw_vwritef(rw_buf *buf, const char *format, va_list ap)
{
    char *formatted;
    int result;

    if (vasprintf(&formatted, format, ap) < 0)
    {
	deblog_error(_("Unable to format string for sending\n"));
    }
    result = rw_write(buf, formatted);
    free(formatted);
    return result;
}

