/* $Id: server.c,v 2.4 2001/10/25 18:16:15 tony Exp $ */

#include "common.h"

#include <limits.h>

#include <dirent.h>
#include <netdb.h>
#include <arpa/inet.h>
#include <wordexp.h>

#include "group.h"
#include "parserc.h"
#include "server.h"
#include "stats.h"

#ifndef MAX_SERVERS
#define MAX_SERVERS 20
#endif
static ServerData *server_table[MAX_SERVERS];

static char *server_newsrc_filename(ServerData *serv)
{
    return XSTRDUPCAT(RC_DIR "/newsrc.", serv->identifier);
}

ServerData *server_new(const char *identifier)
{
    static BOOL in_use_initialised = FALSE;
    static int default_port = -1;
    ServerData *serv;
    int i;

    deblog_info(_("Generating data for server %s\n"), identifier);
    
    if (!in_use_initialised)
    {
	for (i = 0; i < MAX_SERVERS; ++i)
	    server_table[i] = NULL;
	in_use_initialised = TRUE;
    }
    for (i = 0; i < MAX_SERVERS && server_table[i]; ++i);
    if (i >= MAX_SERVERS)
	deblog_error(_("Trying to create too many servers\n"));
    serv = XMALLOC(sizeof(ServerData));
    server_table[i] = serv;
    serv->sidx = i;
    serv->status = server_Unresolved;
    serv->identifier = XSTRDUP(identifier);
    serv->dotquad = NULL;
    if (default_port == -1)
    {
	struct servent *nserv = getservbyname(N_("nntp"), N_("tcp"));
	
	if (nserv)
	{
	    default_port = nserv->s_port;
	    deblog_debug(_("Looked up nntp port: %d\n"), ntohs(default_port));
	}
	else
	{
	    default_port = htons(119);
	    deblog_critical(_("Can't find nntp port in services, using 119\n"));
	}
    }
    serv->port = default_port;
    serv->address = serv->user = serv->password = NULL;
    serv->user = serv->password = NULL;
    if (maincf_get_path())
	serv->path = XSTRDUP(maincf_get_path());
    else
	serv->path = NULL;
    serv->mode_reader = TRUE;
    serv->enter_group = FALSE;
    serv->pipeline = maincf_get_pipeline();
    serv->max_threads = maincf_get_threads_per_server();
    serv->nthreads = 0;
    serv->next_group_stats = serv->group_ix = 0;
    serv->ngroups = serv->group_size = 0;
    serv->group = NULL;
    serv->server_weight = 0;
    serv->feed = TRUE;
    serv->outgoing_size = serv->noutgoing = serv->nuploaded = 0;
    serv->outgoing = NULL;
    deblog_info(_("Server %s has index %d\n"), identifier, i);

    return serv;
}

void server_delete(ServerData *serv)
{
    long int n;
    
    deblog_debug(_("Deleting server data for %s\n"), serv->identifier);
    server_table[serv->sidx] = NULL;
    free(serv->identifier);
    free(serv->dotquad);
    free(serv->address);
    for (n = 0; n < serv->ngroups; ++n)
	group_delete(serv->group[n]);
    free(serv->group);
    for (n = 0; n < serv->noutgoing; ++n)
	free(serv->outgoing[n]);
    free(serv->outgoing);
    free(serv);
}

void server_add_group(ServerData *serv, GroupData *group)
{
    if (serv->ngroups >= serv->group_size)
    {
	serv->group = XREALLOC(serv->group, sizeof(GroupData *) *
		(serv->group_size = serv->group_size ? serv->group_size * 2 :
		 4));
    }
    deblog_debug(_("Group %s added to server %s at index %ld\n"),
	    group->name, serv->identifier, serv->ngroups);
    serv->group[serv->ngroups++] = group;
}

/* NB filename is used in situ, not duplicated */
static void server_add_outgoing(ServerData *serv, char *filename)
{
    if (serv->noutgoing >= serv->outgoing_size)
    {
	serv->outgoing = XREALLOC(serv->outgoing, sizeof(char *) *
		(serv->outgoing_size = serv->outgoing_size ?
		 serv->outgoing_size * 2 : 4));
    }
    deblog_debug(_("Outgoing %s added to server %s at index %ld\n"),
	    filename, serv->identifier, serv->noutgoing);
    serv->outgoing[serv->noutgoing++] = filename;
    stats_inc_total_to_upload();
}

static void server_scan_outgoing_dir(ServerData *serv)
{
    DIR *dp;
    struct dirent *entry;
    char *dirname = XSTRDUPCAT(OUTGOING_DIR "/newsstar/", serv->identifier);

    if ((dp = opendir(dirname)) == NULL)
    {
	deblog_warning(_("Can't open '%s' outgoing directory for server %s\n"),
		dirname, serv->identifier);
	serv->feed = FALSE;
    }
    while ((entry = readdir(dp)) != NULL)
    {
	char *filename;
	
	if (entry->d_name[0] == '.')
	    continue;
	filename = XMALLOC(strlen(dirname) + strlen(entry->d_name) + 2);
	sprintf(filename, "%s/%s", dirname, entry->d_name);
	server_add_outgoing(serv, filename);
	/* Don't free filename, see server_add_outgoing() */
    }
    closedir(dp);
    free(dirname);
}

static const char *server_parserc_handler(const char *group_name, const char
	*valstr, void *dh)
{
    long int last_downloaded = 0;
    ServerData *serv = dh;
    long int n;

    for (n = 0; n < serv->ngroups; ++n)
    {
	if (!strcmp(serv->group[n]->name, group_name))
	{
	    deblog_critical(_("Duplicate entry for group %s on server %s\n"),
		    group_name, serv->identifier);
	    return NULL;
	}
    }
    if (valstr && valstr[0])
	last_downloaded = strtol(valstr, NULL, 10);
    server_add_group(serv, group_new(group_name, last_downloaded));
    return NULL;
}

void server_read_newsrc(ServerData *serv)
{
    char *filename;

    deblog_info(_("Reading newsrc file for %s\n"), serv->identifier);
    filename = server_newsrc_filename(serv);
    parserc_register_master_handler(server_parserc_handler, serv);
    parserc_read_named(filename, deblog_Level_Error);
    free(filename);
    serv->status = server_Unresolved;
}

void server_write_newsrc(ServerData *serv)
{
    char *filename;
    FILE *fp;
    long int n;

    if (!serv->ngroups)
    {
	deblog_warning(_("Skipping write of newsrc file for %s\n"),
		serv->identifier);
	return;
    }
    deblog_info(_("Writing newsrc file for %s\n"), serv->identifier);
    filename = server_newsrc_filename(serv);
    fp = fopen(filename, N_("w"));
    if (!fp)
    {
	deblog_critical(_("Can't open %s for writing\n"), filename);
	free(filename);
	return;
    }
    for (n = 0; n < serv->ngroups; ++n)
	group_write_newsrc(serv->group[n], fp);
    fclose(fp);
    free(filename);
}

void server_read_cf(ServerData *serv)
{
    char *filename;
    long int port = -1;
    char *path = NULL;

    deblog_info(_("Reading cf file for %s\n"), serv->identifier);
    filename = XSTRDUPCAT(CONF_DIR "/cf.", serv->identifier);
    parserc_register_string(N_("server"), &serv->address);
    parserc_register_string(N_("path"), &path);
    parserc_register_int(N_("port"), &port);
    parserc_register_string(N_("user"), &serv->user);
    parserc_register_string(N_("password"), &serv->password);
    parserc_register_bool(N_("feed"), &serv->feed);
    parserc_register_bool(N_("mode_reader"), &serv->mode_reader);
    parserc_register_bool(N_("enter_group"), &serv->enter_group);
    parserc_register_bool(N_("pipeline"), &serv->pipeline);
    parserc_register_int(N_("max_threads"), &serv->max_threads);
    parserc_register_int(N_("server_weight"), &serv->server_weight);
    parserc_read_named(filename, deblog_Level_Info);
    free(filename);
    if (port != -1)
	serv->port = htons(port);
    if (path)
    {
	if (serv->path)
	    free(serv->path);
	serv->path = path;
    }
}

static void server_exit_handler(void)
{
    static BOOL recursive_exit = FALSE;

    deblog_debug(_("Entering server_exit_handler\n"));
    if (!recursive_exit)
    {
	int n;

	recursive_exit = TRUE;
	for (n = 0; n < MAX_SERVERS; ++n)
	{
	    if (!server_table[n])
		continue;
	    server_write_newsrc(server_table[n]);
	}
    }
    deblog_debug(_("Leaving server_exit_handler\n"));
}

void server_scan_rc_dir(void)
{
    wordexp_t we;
    unsigned int n;
    long int max_threads = 0;

    we.we_wordc = 0;
    we.we_wordv = 0;
    we.we_offs = 0;
    if ((n = wordexp(RC_DIR "/newsrc.*", &we, 0)) != 0)
    {
	deblog_error(_("Expansion of newsrc wildcard failed (code %d)\n"), n);
    }
    if (we.we_wordc == 1 && !strcmp(we.we_wordv[0], RC_DIR "/newsrc.*"))
    {
	deblog_error(_("No newsrc files\n"));
    }
    for (n = 0; n < we.we_wordc; ++n)
    {
	ServerData *serv;
	
	deblog_debug(_("wordexp found `%s'\n"), we.we_wordv[n]);
	if (!strcmp(we.we_wordv[n], RC_DIR "/newsrc.sample"))
	{
	    deblog_warning(
	      _("newsrc.sample file found; you should delete or rename it\n"));
	    continue;
	}
	serv = server_new(we.we_wordv[n] + sizeof(RC_DIR "/newsrc.") - 1);
	server_read_newsrc(serv);
	server_read_cf(serv);
	max_threads += serv->max_threads;
	if (!serv->address)
	    serv->address = XSTRDUP(serv->identifier);
	if (serv->feed)
	    server_scan_outgoing_dir(serv);
    }
    wordfree(&we);
    atexit(server_exit_handler);
    if (max_threads < maincf_get_max_threads())
	maincf_reset_max_threads(max_threads);
}

long int server_get_next_group_stats(ServerData *serv)
{
    if (serv->next_group_stats >= serv->ngroups)
	return -1;
    serv->group[serv->next_group_stats]->status = group_Allocated;
    return serv->next_group_stats++;
}

long int server_get_next_group_articles(ServerData *serv)
{
    while (++serv->group_ix < serv->ngroups)
    {
	if (serv->group[serv->group_ix]->status != group_AllDone)
	    return serv->group_ix;
    }
    return -1;
    return serv->group_ix;
}

/* Minimum number of articles worth starting a new thread for */
#ifndef MIN_WORTH_GETTING
#define MIN_WORTH_GETTING 25
#endif

server_GFT_Result server_get_for_threading(ServerData **serv_out)
{
    ServerData *best_server = NULL;
    long int best_score = MIN_WORTH_GETTING;
    int n;
    BOOL all_done = TRUE;

    /* Find most eligible server */
    deblog_info(
	    _("gft: Finding most eligible server to start a thread for\n"));
    for (n = 0; n < MAX_SERVERS; ++n)
    {
	long int score = 0;
	ServerData *serv;
	BOOL all_unknown;
	long int group_ix;

	serv = server_table[n];
	if (!serv)
	    continue;
	deblog_info(_("gft: %s has %d/%d threads allocated\n"),
		serv->identifier, serv->nthreads, serv->max_threads);
	deblog_info(_("gft: %s has had %d/%d groups looked at\n"),
		serv->identifier, serv->next_group_stats, serv->ngroups);
	if (serv->nthreads >= serv->max_threads)
	{
	    /* This is basically equivalent to AllDone */
	    score = -2;
	    continue;
	}
	switch (serv->status)
	{
	    case server_Error:
		deblog_info(_("gft: %s has an error condition\n"),
			serv->identifier);
		score = -2;
		break;
	    case server_AllDone:
		deblog_info(_("gft: %s all done\n"),
			serv->identifier);
		score = -2;
		break;
	    case server_Resolving:
		deblog_info(_("gft: Still resolving address for %s\n"),
			serv->identifier);
		score = -2;
		all_done = FALSE;
		break;
	    case server_Unresolved:
		/* Great, we really want to connect to this one */
		deblog_info(_("gft: %s has unknown address\n"),
			serv->identifier);
		score += 20000;
		all_done = FALSE;
		break;
	    case server_Resolved:
		/* Great, we really want to connect to this one */
		deblog_info(
	    _("gft: %s has known address but hasn't been connected to yet\n"),
			serv->identifier);
		score += 10000;
		all_done = FALSE;
		break;
	    case server_Connected:
		/* Might have some info available about number of articles
		 * available we can use to influence score */
		deblog_info(_("gft: %s has already been connected to\n"),
			serv->identifier);
		all_done = FALSE;
		/* all_unknown is only used by following loop */
		all_unknown = TRUE;
		for (group_ix = 0; group_ix < serv->next_group_stats;
			++group_ix)
		{
		    GroupData *group = serv->group[group_ix];

		    switch (group->status)
		    {
			case group_Downloading:
			    score += group->last_available -
				group->last_requested;
			    deblog_info(
				_("gft: %s on %s has %d %s available\n"),
				group->name, serv->identifier,
				group->last_available - group->last_requested,
				group->last_available - group->last_requested
				    == 1 ?
				    _("article") : _("articles"));
			    all_unknown = FALSE;
			    break;
			case group_AllDone:
			    deblog_info(
				_("gft: %s on %s has no articles available\n"),
				group->name, serv->identifier);
			    break;
			default:
			    deblog_info(
				_("gft: No stats about %s on %s available\n"),
				group->name, serv->identifier);
			    break;
		    }
		} /* for (group_ix...) */
		if (all_unknown)
		{
		    deblog_info(_("gft: %s has no stats available\n"),
			    serv->identifier);
		    score = -2;
		}
		break;

	} /* switch (serv->status) */
	if (score >= MIN_WORTH_GETTING)
	{
	    score += serv->server_weight;
	    /* Take number of threads into account */
	    score += 1000 * (serv->max_threads - serv->nthreads);
	    if (!serv->nthreads)
	    {
		/* Give it a boost */
		score += 1000;
	    }
	}
	deblog_info(_("gft: %s has scored %ld (best %ld)\n"),
		serv->identifier, score, best_score);
	if (score > best_score)
	{
	    best_server = serv;
	    best_score = score;
	}
    } /* for (n = 0; n < MAX_SERVERS; ++n) */
    if (best_server)
    {
	deblog_info(_("gft: %s has best score with %ld\n"),
		best_server->identifier, best_score);
	*serv_out = best_server;
	return server_GFT_OK;
    }
    deblog_info(all_done ? _("gft: All done\n") : _("All busy\n"));
    return all_done ? server_GFT_AllDone : server_GFT_AllBusy;
}

