/* $Id: stats.c,v 2.2 2001/10/23 19:03:27 tony Exp $ */

#include "common.h"

#include <limits.h>
#include <stdio.h>

#include "deblog.h"
#include "nscurses.h"
#include "stats.h"

static struct timeval stats_start_time;
static long int stats_current_dups = 0;
static long int stats_history_dups = 0;
static long int stats_total_bytes_rcvd = 0;
static long int stats_total_bytes_sent = 0;

static long int stats_total_to_upload = 0;
static long int stats_uploaded = 0;
static long int stats_duplicates = 0;
static long int stats_bad_uploads = 0;

/* A node on a list for holding last few packets of info about bytes
 * transferred */
typedef struct stats_node {
    struct stats_node *next;
    struct timeval tv;
    long int bytes;
} stats_node;

static stats_node *stats_sent_head = NULL;
static stats_node *stats_rcvd_head = NULL;
static stats_node *stats_sent_tail = NULL;
static stats_node *stats_rcvd_tail = NULL;

void stats_start(void)
{
    struct timezone tz;

    if (gettimeofday(&stats_start_time, &tz))
	deblog_error(_("Can't read time\n"));
}

/* Get bytes per sec transferred over last 5 seconds */
static long int stats_get_bps(stats_node **head, stats_node **tail)
{
    /* Floating point because intermediate calculations can get out of range
     * if we try to stick to ints (even long) */
    double total = 0.0;
    stats_node *node;
    struct timeval now, difference, since = { 0, 0 };
    struct timezone tz;
    static struct timeval period = { 5, 0 };
    BOOL first = TRUE;

    gettimeofday(&now, &tz);
    for (node = *head; node; )
    {
	stats_node *next = node->next;

	xtimeval_subtract(&now, &node->tv, &difference);
	if (timercmp(&difference, &period, > ))
	{
	    /* Items should only be removed from head of list unless we've gone
	     * drastically wrong */
	    *head = node->next;
	    if (*tail == node)
		*tail = NULL;
	    free(node);
	}
	else
	{
	    if (first)
	    {
		first = FALSE;
		since = node->tv;
	    }
	    total += (double) node->bytes;
	}
	node = next;
    }
    if (first)
	return 0;
    xtimeval_subtract(&now, &since, &difference);
    return (long int) (total / ((double) difference.tv_sec + (double)
		difference.tv_usec / 1E6));
}

static void stats_queue_add(stats_node **head, stats_node **tail,
	const struct timeval *tv, long int bytes)
{
    stats_node *node = XMALLOC(sizeof(stats_node));

    node->next = NULL;
    node->tv = *tv;
    node->bytes = bytes;
    if (!*head)
	*head = node;
    if (*tail)
	(*tail)->next = node;
    *tail = node;
}

/* STATS_INTERVAL is interval between refreshes of info on screen in ms */
#ifdef STATS_INTERVAL
#define STATS_INTERVAL_TV \
    { (STATS_INTERVAL) / 1000, ((STATS_INTERVAL) * 1000) % 1000000 }
#else
#define STATS_INTERVAL_TV { 0, 350000 }
#endif
static struct timeval stats_interval = STATS_INTERVAL_TV; 

static long int stats_total_to_fetch = 0;
static long int stats_fetched = 0;

void stats_display(void)
{
    char sent[16];
    char rcvd[16];

    strcpy(rcvd, XNTOKILOMEG(stats_get_bps(&stats_rcvd_head,
		    &stats_rcvd_tail)));
    strcpy(sent, XNTOKILOMEG(stats_get_bps(&stats_sent_head,
		    &stats_sent_tail)));
    if (nscurses_active())
    {
	nscurses_stats(stats_fetched, stats_total_to_fetch, rcvd,
		stats_uploaded, stats_total_to_upload, sent);
    }
    else
    {
	printf(_("\r  << %ld/%ld @ %sB/s \t >> %ld/%ld @ %sB/s \t\t"),
		stats_fetched, stats_total_to_fetch, rcvd,
		stats_uploaded, stats_total_to_upload, sent);
	fflush(stdout);
	deblog_need_lf();
    }
}

static void stats_update(void)
{
    static BOOL first = TRUE;
    static struct timeval last_time = {0, 0};
    struct timeval now;
    struct timeval difference;
    struct timezone tz;

    if (gettimeofday(&now, &tz))
	deblog_error(_("Can't read time\n"));
    /* Subtract last_time from now (doesn't matter about last_time being
     * meaningless if first) */
    xtimeval_subtract(&now, &last_time, &difference);
    if (first || timercmp(&difference, &stats_interval, > ))
    {
	first = FALSE;
	last_time = now;
	stats_display();
    }
}

void stats_inc_total_to_fetch(void)
{
    ++stats_total_to_fetch;
    stats_update();
}

void stats_dec_total_to_fetch(void)
{
    --stats_total_to_fetch;
    stats_update();
}

void stats_inc_fetched(void)
{
    ++stats_fetched;
    stats_update();
}

void stats_update_bytes_rcvd(long int bytes, const struct timeval *tv)
{
    /* deblog_debug(_("%ld bytes rcvd since %.2f\n"), bytes, FLOAT_TV(tv)); */
    stats_queue_add(&stats_rcvd_head, &stats_rcvd_tail, tv, bytes);
    stats_total_bytes_rcvd += bytes;
    stats_update();
}

void stats_update_bytes_sent(long int bytes, const struct timeval *tv)
{
    /* deblog_debug(_("%ld bytes sent since %.2f\n"), bytes, FLOAT_TV(tv)); */
    stats_queue_add(&stats_sent_head, &stats_sent_tail, tv, bytes);
    stats_total_bytes_sent += bytes;
    stats_update();
}

void stats_inc_current_dups(void)
{
    ++stats_current_dups;
}

void stats_inc_history_dups(void)
{
    ++stats_history_dups;
}

void stats_inc_total_to_upload(void)
{
    ++stats_total_to_upload;
}

void stats_dec_total_to_upload(void)
{
    --stats_total_to_upload;
    stats_update();
}

void stats_inc_uploaded(void)
{
    ++stats_uploaded;
    stats_update();
}

void stats_inc_duplicates(void)
{
    ++stats_duplicates;
}

void stats_inc_bad_uploads(void)
{
    ++stats_bad_uploads;
}

void stats_summary(void)
{
    struct timeval now, difference;
    struct timezone tz;
    char sent[16];
    char rcvd[16];

    if (gettimeofday(&now, &tz))
	deblog_error(_("Can't read time\n"));
    xtimeval_subtract(&now, &stats_start_time, &difference);

    deblog_message(_("%ld %s downloaded\n"), stats_fetched,
	    stats_fetched == 1 ? _("article") : _("articles"));
    deblog_message(_("%ld %s skipped, %ld already in history\n"),
	    stats_current_dups,
	    stats_current_dups == 1 ? _("duplicate") : _("duplicates"),
	    stats_history_dups);
    deblog_message(_("%ld %s uploaded\n"), stats_uploaded,
	    stats_uploaded == 1 ? _("article") : _("articles"));
    deblog_message(
	    _("%ld %s rejected as duplicates, %ld for other reasons\n"),
	    stats_duplicates,
	    stats_duplicates == 1 ? _("upload") : _("uploads"),
	    stats_bad_uploads);
    strcpy(rcvd, XNTOKILOMEG(stats_total_bytes_rcvd));
    strcpy(sent, XNTOKILOMEG(stats_total_bytes_sent));
    deblog_message(_("%sB received, %sB sent in %.2f secs\n"), rcvd, sent,
	    FLOAT_TV(&difference));
    strcpy(rcvd, XNTOKILOMEG((long int) ((double) stats_total_bytes_rcvd /
		    ((double) difference.tv_sec + (double) difference.tv_usec /
		     1E6))));
    strcpy(sent, XNTOKILOMEG((long int) ((double) stats_total_bytes_sent /
		    ((double) difference.tv_sec + (double) difference.tv_usec /
		     1E6))));
    deblog_message(_("(%sB/s, %sB/s)\n"), rcvd, sent);
}
