/* $Id: stdx.c,v 2.0 2001/07/26 16:35:12 tony Exp $ */

#include "common.h"

#define XBUFLEN 200

static void stdx_log_strerror(const char *msg, va_list ap)
{
    char *formatted, *complete;

    if (vasprintf(&formatted, msg, ap) < 0 
	    || asprintf(&complete, N_("%s: %s\n"), formatted, strerror(errno))
		< 0)
    {
	fprintf(stderr, _("Can't generate meaningful error message\n"));
	exit(1);
    }
    free(formatted);
    fprintf(stderr, complete);
    free(complete);
    exit(1);
}

void *xmalloc(size_t sz, const char *msg, ...)
{
    void *blk = malloc(sz);
    
    if (!blk)
    {
	va_list ap;

	va_start(ap, msg);
	vfprintf(stderr, msg, ap);
	va_end(ap);
	exit(1);
    }
    return blk;
}
    
void *xrealloc(void *old, size_t sz, const char *msg, ...)
{
    void *blk = realloc(old, sz);
    if (!blk)
    {
	va_list ap;

	va_start(ap, msg);
	vfprintf(stderr, msg, ap);
	va_end(ap);
	exit(1);
    }
    return blk;
}

char *xstrdup(const char *s, const char *msg, ...)
{
    char *news = strdup(s);

    if (!news)
    {
	va_list ap;

	va_start(ap, msg);
	vfprintf(stderr, msg, ap);
	va_end(ap);
	exit(1);
    }
    return news;
}

char *xstrdupcat(const char *s1, const char *s2, const char *msg, ...)
{
    char *news = malloc(strlen(s1) + strlen(s2) + 1);

    if (!news)
    {
	va_list ap;

	va_start(ap, msg);
	vfprintf(stderr, msg, ap);
	va_end(ap);
	exit(1);
    }
    strcpy(news, s1);
    strcat(news, s2);
    return news;
}

int xpipe(int fd[2], const char *msg, ...)
{
    va_list ap;

    if (!pipe(fd))
	return 0;
    va_start(ap, msg);
    stdx_log_strerror(msg, ap);
    va_end(ap);
    exit(1);
    return -1;
}

void xtimeval_subtract(struct timeval *higher, struct timeval *lower, struct
	timeval *result)
{
    result->tv_sec = higher->tv_sec - lower->tv_sec;
    if (higher->tv_usec < lower->tv_usec)
    {
	--result->tv_sec;
	result->tv_usec = higher->tv_usec + 1000000 - lower->tv_usec;
    }
    else
    {
	result->tv_usec = higher->tv_usec - lower->tv_usec;
    }
}

const char *xntokilomeg(long int n, char suffix)
{
    static char result[16];
    long int rem = 0;
    float f;

    if (suffix)
    {
	rem = n % 1000;
	n /= 1000;
    }
    if (n > 9999 || n < -9999)
    {
	switch (suffix)
	{
	    case 0:
		return xntokilomeg(n, 'K');
	    case 'K':
		return xntokilomeg(n, 'M');
	    case 'M':
		return xntokilomeg(n, 'G');
	    case 'G':
		return xntokilomeg(n, 'T');
	    case 'T':
		return xntokilomeg(n, 'P');
	}
    }
    f = (float) n + (float) rem / 1000.0F;
    if (n >= 1000 || n <= -1000 || !suffix)
	sprintf(result, "%ld%c", n, suffix);
    else if (n >= 100 || n <= -100)
	sprintf(result, "%.1f%c", f, suffix);
    else 
	sprintf(result, "%.2f%c", f, suffix);
    return result;
}

char *xtempnam(const char *dir, const char *pfx, const char *msg, ...)
{
    char *result = tempnam(dir, pfx);
    if (!result)
    {
	va_list ap;

	va_start(ap, msg);
	stdx_log_strerror(msg, ap);
	va_end(ap);
	exit(1);
    }
    return result;
}


/* Rest not yet implemented and need fixing */
#if 0
FILE *xfopen(const char *path, const char *mode, const char *msg, ...)
{
    FILE *fp = fopen(path, mode);
    if (!fp)
    {
	char buf[XBUFLEN];
	va_list ap;
	va_start(ap, msg);
	display_endwin();
	VSNPRINTF(buf, XBUFLEN, msg, ap);
	perror(buf);
	va_end(ap);
	exit(1);
    }
    return fp;
}

char *xfgets(char *buf, int size, FILE *fp, const char *msg, ...)
{
    char *fgres = fgets(buf, size, fp);
    if (!fgres && ferror(fp))
    {
	char buf2[XBUFLEN];
	va_list ap;
	va_start(ap, msg);
	display_endwin();
	VSNPRINTF(buf2, XBUFLEN, msg, ap);
	perror(buf2);
	va_end(ap);
	exit(1);
    }
    return fgres;
}

void xchdir(const char *dir, const char *msg, ...)
{
    if (chdir(dir) == -1)
    {
	char buf[XBUFLEN];
	va_list ap;
	va_start(ap, msg);
	display_endwin();
	VSNPRINTF(buf, XBUFLEN, msg, ap);
	perror(buf);
	va_end(ap);
	exit(1);
    }
}

#define FCOPY_BUFSIZE 4096

int xfcopy(const char *src, const char *dest, const char *msg, ...)
{
    FILE *fpin = 0;
    FILE *fpout = 0;
    char buffer[FCOPY_BUFSIZE];
    int success = 1;
    fpin = fopen(src, N_("r"));
    fpout = fopen(dest, N_("w"));
    if (!fpin || ! fpout)
	success = 0;
    while (success && !feof(fpin))
    {
	int n = fread(buffer, sizeof(char), FCOPY_BUFSIZE, fpin);
	if (ferror(fpin))
	{
	    success = 0;
	}
	else
	{
	    fwrite(buffer, sizeof(char), n, fpout);
	    if (ferror(fpout))
		success = 0;
	}
    }
    if (fpin)
	fclose(fpin);
    if (fpout)
	fclose(fpout);
    if (!success)
    {
	char buf[XBUFLEN];
	va_list ap;
	va_start(ap, msg);
	input_delwin();
	display_endwin();
	VSNPRINTF(buf, XBUFLEN, msg, ap);
	perror(buf);
	va_end(ap);
    }
    return success;
}

int xrename(const char *src, const char *dest, const char *msg, ...)
{
    int result = rename(src, dest);
    if (result == -1)
    {
	char buf[XBUFLEN];
	va_list ap;
	va_start(ap, msg);
	input_delwin();
	display_endwin();
	VSNPRINTF(buf, XBUFLEN, msg, ap);
	perror(buf);
	va_end(ap);
    }
    return result;
}

int xremove(const char *filename, const char *msg, ...)
{
    int result = remove(filename);
    if (result == -1)
    {
	char buf[XBUFLEN];
	va_list ap;
	va_start(ap, msg);
	input_delwin();
	display_endwin();
	VSNPRINTF(buf, XBUFLEN, msg, ap);
	perror(buf);
	va_end(ap);
    }
    return result;
}

#endif /* commented block */
